import re, StringIO, urllib, time, datetime
import Image, Geocode, Store
import BeautifulSoup
import PIL.Image, PIL.ImageDraw
from twisted.internet import reactor, defer
from twisted.web import client

class DumbClient(client.HTTPPageGetter):
    """ Single-request client that knows to call its factory
        deferred with more than just a string of page content.
    """
    def handleResponse(self, response):
        client.HTTPPageGetter.handleResponse(self, response)
        self.response = response
        if hasattr(self.factory, 'agentDeferred'):
            self.factory.agentDeferred.callback(self)
    
class DumbClientFactory(client.HTTPClientFactory):
    """ Single-request client factory that has a secondary
        callback for use by DumbClient.handleResponse.
    """
    protocol = DumbClient
    def setResponseCallback(self, callback):
        self.agentDeferred = defer.Deferred()
        self.agentDeferred.addCallback(callback)

class MapRequest:
    
    def __init__(self, areaType, areaName, dateStart, dateEnd, crimeType):
        self.areaType  = areaType
        self.areaName  = areaName
        self.dateStart = dateStart
        self.dateEnd   = dateEnd
        self.crimeType = crimeType
        
        self.cookieJar = {}
    
    def __str__(self):
        return '%(areaName)s, %(crimeType)s: %(dateStart)s - %(dateEnd)s' % self.__dict__
        
    def setIDs(self, appID, sessID):
        self.appID  = appID
        self.sessID = sessID

class CrimeRequest:
    
    def __init__(self, crimeMatch, mapRequest, district):
        self.crimeMatch = crimeMatch
        self.mapRequest = mapRequest
        self.district   = district
        self.next       = None
        self.remaining  = 0
        
    def __str__(self):
        crimeName      = self.crimeMatch.crime.name
        crimeX, crimeY = self.crimeMatch.center()
        mapRequestStr  = str(self.mapRequest)
        remaining      = self.remaining
        
        return '%(crimeName)s @%(crimeX)d,%(crimeY)d (%(mapRequestStr)s), %(remaining)d remaining' % locals()
        
    def nextRequest(self, request=None):
        if request is None:
            return self.next
            
        else:
            if self.next is not None:
                # chain 'em up
                request.nextRequest(self.next)

            self.next = request
            self.remaining = request.remaining + 1

class CrimeAgent:
    userAgent = 'Mozilla/5.0 (Macintosh; U; PPC Mac OS X Mach-O; en-US; rv:1.8.0.8) Gecko/20061025 Firefox/1.5.0.8'
    
    def __init__(self, dataStore, dateStart, dateEnd, areaName=None, crimeType=None):
        self.requestQueue = []
        self.requestInterval = 5
        self.reaperInterval = 5 * 60
        self.requestsInProgress = 0
        self.completedRequests = 0
        
        self.dataStore = dataStore
        self.dateStart = dateStart
        self.dateEnd   = dateEnd
        
        if areaName is None:
            self.areaType = 'Council_District_Renderer'
            self.areaNames = Geocode.districts().keys()
        else:
            self.areaType = 'Council_District_Renderer'
            self.areaNames = [areaName]

        if crimeType is None:
            self.crimeTypes = Image.crimes().keys()
        else:
            self.crimeTypes = [crimeType]
        
    def requestStatus(self):
        return 'Requests: %d in queue, %d in progress, %d completed' % (len(self.requestQueue), self.requestsInProgress, self.completedRequests)
    
    def reapRequests(self, lastCompletedRequests=None):
        if lastCompletedRequests == self.completedRequests:
            print '-' * 64
            print 'Finished:', self.reaperInterval, 'seconds have elapsed, completed requests still at', lastCompletedRequests
            reactor.stop()
        
        else:
            reactor.callLater(self.reaperInterval, self.reapRequests, self.completedRequests)
        
    def makeRequest(self, callback, *args, **kwargs):
        factory = DumbClientFactory(*args, **kwargs)
        factory.setResponseCallback(callback)
        reactor.connectTCP(factory.host, 80, factory)
        
    def queueRequests(self):
        for areaName in self.areaNames:
            for crimeType in self.crimeTypes:
                self.requestQueue.insert(0, MapRequest(self.areaType, areaName,
                                                       self.dateStart, self.dateEnd,
                                                       crimeType))

        self.makeNextRequest()
        self.reapRequests()

    def makeNextRequest(self):
        if len(self.requestQueue):
            request = self.requestQueue.pop()
            self.requestsInProgress += 1
            
            if request.__class__ is MapRequest:
                reactor.callLater(0, self.enterSite, request)
        
            elif request.__class__ is CrimeRequest:
                reactor.callLater(0, self.requestCrimeDetail, request)

        reactor.callLater(self.requestInterval, self.makeNextRequest)
        
    def enterSite(self, mapRequest):
        print '-' * 64
        print 'Entering site for request:', str(mapRequest)
        print self.requestStatus()

        agent = self

        def onEnterSite(client):
            agent.onEnterSite(client, mapRequest)

        self.makeRequest(onEnterSite,
                         'http://gismaps.oaklandnet.com/crimewatch/default.asp',
                         cookies=mapRequest.cookieJar, agent=self.userAgent)
        
    def onEnterSite(self, client, mapRequest):
        print '-' * 64
        print 'Site entered:', str(mapRequest)
        self.acceptTerms(mapRequest)
        
    def acceptTerms(self, mapRequest):
        print '-' * 64
        print 'Accepting terms for request:', str(mapRequest)
        print self.requestStatus()

        agent = self

        def onAcceptTerms(client):
            agent.onAcceptTerms(client, mapRequest)
    
        self.makeRequest(onAcceptTerms,
                         'http://gismaps.oaklandnet.com/crimewatch/default.asp',
                         method='POST', postdata=urllib.urlencode({'fldAccept': 'yes'}),
                         headers={'Content-Type': 'application/x-www-form-urlencoded'},
                         cookies=mapRequest.cookieJar, agent=self.userAgent)
    
    def onAcceptTerms(self, client, mapRequest):
        print '-' * 64
        print 'Terms accepted:', str(mapRequest)
        self.requestMapFrame(mapRequest)
        
    def getMapFrameRequestParams(self, mapRequest):
        """ Get a dictionary of map request parameters.
        """
        dateStart = mapRequest.dateStart
        dateEnd   = mapRequest.dateEnd
        crimeType = mapRequest.crimeType
        
        return {'mapfunction1':           51,
                'mapFunction2':           40,
                'maplayeractivebyid2':    '',
                'maplayervisiblebyid2':   '',
                'queryName2':             mapRequest.areaType,
                'queryValue2':            mapRequest.areaName,
                'bufferDistance3':        0,
                'bufferUnits3':           0,
                'bufferSource3':          2,
                'bufferTarget3':          0,
                'mapFunction3':           25,
                'mapFunction4':           34,
                'whereExpr4':             "((CV_DATE>=%(dateStart)s) AND (CV_DATE<=%(dateEnd)s)) AND (CV_LEGEND IN ('%(crimeType)s'))" % locals(),
                'mapFunction5':           76,
                'mapFunction6':           37,
                'maplayeractivebyid7':    'incidents',
                'maplayervisiblebyid7':   'incidents',
                'windowwidth':            Geocode.districts()[mapRequest.areaName].width,
                'windowheight':           Geocode.districts()[mapRequest.areaName].height}
    
    def requestMapFrame(self, mapRequest):
        """ Request HTML source of map frame.
        """
        print '-' * 64
        print 'Sending map frame request:', str(mapRequest)
        print self.requestStatus()

        params = self.getMapFrameRequestParams(mapRequest)
        agent = self
        
        def onRequestMapFrame(client):
            reactor.callLater(0, agent.onRequestMapFrame,
                                 client.factory.url, client.response,
                                 mapRequest)
                  
        self.makeRequest(onRequestMapFrame,
                         'http://gismaps.oaklandnet.com/crimewatch/map.asp?'+urllib.urlencode(params),
                         headers={'Referer': 'http://gismaps.oaklandnet.com/crimewatch/wizard.asp'},
                         cookies=mapRequest.cookieJar, agent=self.userAgent)
    
    def onRequestMapFrame(self, mapFrameURL, mapHTML, mapRequest):
        """ HTML source of map frame has been retrieved.
        """
        print '-' * 64
        print 'Map frame HTML:', len(mapHTML), 'bytes'
        print 'Request:', str(mapRequest)
        
        # the map JPEG url is hidden inside a div with id="MapImg"
        mapSoup = BeautifulSoup.BeautifulSoup(mapHTML)
        mapDiv = mapSoup.first('div', {'id': 'MapImg'})
        mapURL = re.search(r'\bBackground-image:\s*url\((http://[^\)]+)\);', mapDiv['style'], re.I).group(1)
        
        # for use later
        mapRequest.setIDs(mapSoup.first('input', {'type': 'hidden', 'name': 'AppID'})['value'],
                          mapSoup.first('input', {'type': 'hidden', 'name': 'SessID'})['value'])

        self.requestMapImage(mapURL, mapFrameURL, mapRequest)
        
    def requestMapImage(self, url, referer, mapRequest):
        """ Request raw map image data.
        """
        agent = self
        
        def onRequestMapImage(client):
            reactor.callLater(0, agent.onRequestMapImage,
                                 PIL.Image.open(StringIO.StringIO(client.response)),
                                 mapRequest)
                  
        self.makeRequest(onRequestMapImage,
                         url, headers={'Referer': referer},
                         cookies=mapRequest.cookieJar, agent=self.userAgent)

    def onRequestMapImage(self, mapImage, mapRequest):
        """ Raw map image data has been retrieved.
        """
        print '-' * 64
        print 'Map image size:', mapImage.size
        print 'Request:', str(mapRequest)
        
        self.extractImageFeatures(mapImage, mapRequest)

    def extractImageFeatures(self, mapImage, mapRequest):
        """ Extract useful features from raw map image data.
        """
        try:
            # separate channels of original image, as arrays
            mapImageR, mapImageG, mapImageB = [Image.img2arr(chan) for chan in mapImage.split()]
            
            crime = Image.crimes().get(mapRequest.crimeType)
            
            print 'Isolating', crime.name, 'pixels...'
            mask = Image.getMask(mapImage, mapImageR, mapImageG, mapImageB, crime)
            #mask.show()
            
            print '  Checking', mask.histogram()[-1], 'of', (mask.size[0] * mask.size[1]), 'total pixels for matches...'
            for match in Image.getMatches(mapImage, mask, crime):
                box = (match.box[0] - 1, match.box[1] - 1, match.box[2], match.box[3])
                
                if match.part == Image.Icon.WHOLE:
                    PIL.ImageDraw.Draw(mapImage).rectangle(box, outline=(0x33, 0x33, 0x66), fill=None)
    
                else:
                    PIL.ImageDraw.Draw(mapImage).rectangle(box, outline=(0xCC, 0x99, 0x99), fill=None)
                    
                district = Geocode.districts()[mapRequest.areaName]
                nextCrimeRequest = CrimeRequest(match, mapRequest, district)
                nextCrimeRequest.nextRequest(locals().get('crimeRequest'))
                crimeRequest = nextCrimeRequest
    
            if locals().has_key('crimeRequest'):
                self.requestQueue.insert(0, crimeRequest)
                
        except Exception, e:
            print 'Something went wrong while extracting image features:'
            print e
        
        self.finishRequest(mapRequest)
            
        #print '  Showing image...'
        #mapImage.show()
        
    def getCrimeDetailRequestParams(self, crimeRequest):
        """ Get a dictionary of crime detail request parameters.
        """
        crimeMatch = crimeRequest.crimeMatch
        mapRequest = crimeRequest.mapRequest
        
        return {'WinVisible':               0,
                'WinLeft':                  -1,
                'WinTop':                   -1,
                'overviewVisible':          0,
                'overviewLeft':             100,
                'overviewTop':              26,
                'legendVisible':            1,
                'MessageAvailable':         1,
                'messageVisible':           0,
                'messageLeft':              100,
                'messageTop':               100,
                'AttributesAvailable':      0,
                'SummaryReportAvailable':   0,
                'CurrTool':                 20,
                'ToolsVisible':             1,
                'AppID':                    mapRequest.appID,
                'SessID':                   mapRequest.sessID,
                'WindowWidth':              Geocode.districts()[mapRequest.areaName].width,
                'WindowHeight':             Geocode.districts()[mapRequest.areaName].height,
                'MapMinX':                  int(crimeMatch.center()[0]),
                'MapMinY':                  int(crimeMatch.center()[1]),
                'MapMaxX':                  0,
                'MapMaxY':                  0,
                'EnvIsMapUnits':            0,
                'EnvIsDegrees':             0,
                'MouseBtn':                 0,
                'ShiftKey':                 0,
                'MapLayerActive':           'TEST',
                'MapLyrVis_1':              1,
                'MapLyrVis_councildistrict':0,
                'MapLyrVis_2':              0,
                'MapLyrVis_7':              0,
                'MapLyrVis_5':              1,
                'MapLyrVis_4':              0,
                'MapLyrVis_PARKS':          1,
                'MapLyrVis_hydro':          1,
                'MapLyrVis_6':              0,
                'MapLyrVis_City_Street':    1,
                'MapLyrVis_streets':        1,
                'MapLyrVis_Major_Street':   1,
                'MapLyrVis_freeways':       1,
                'MapLyrVis_10':             0,
                'MapLyrVis_11':             0,
                'MapLyrVis_12':             0,
                'MapLyrVis_13':             0,
                'MapLyrVis_14':             0,
                'MapLyrVis_15':             0,
                'MapLyrVis_TEST':           1,
                'MapFunction':              20,
                'XML':                      '',
                'QueryName':                '',
                'QueryValue':               '',
                'BufferSource':             '',
                'BufferTarget':             '',
                'BufferDistance':           '',
                'BufferUnits':              0,
                'BufferDownload':           0,
                'BufferList':               ''}
    
    def requestCrimeDetail(self, crimeRequest):
        """ Request HTML source of crime match detail.
        """
        print '-' * 64
        print 'Sending crime detail request:', str(crimeRequest)
        print self.requestStatus()

        params = self.getCrimeDetailRequestParams(crimeRequest)
        agent = self

        def onRequestCrimeDetail(client):
            reactor.callLater(0, agent.onRequestCrimeDetail,
                                 client.response,
                                 crimeRequest)
                  
        self.makeRequest(onRequestCrimeDetail,
                         'http://gismaps.oaklandnet.com/crimewatch/map.asp',
                         method='POST', postdata=urllib.urlencode(params),
                         headers={'Content-Type': 'application/x-www-form-urlencoded', 'Referer': 'http://gismaps.oaklandnet.com/crimewatch/map.asp'},
                         cookies=crimeRequest.mapRequest.cookieJar, agent=self.userAgent)

    def onRequestCrimeDetail(self, crimeDetailsHTML, crimeRequest):
        """ HTML source of crime match detail has been retrieved.
        """
        print '-' * 64
        print 'Crime details HTML:', len(crimeDetailsHTML), 'bytes'
        print 'Request:', str(crimeRequest)

        # the iframe URL is deep inside a javascript statement
        iframeURL = 'http://gismaps.oaklandnet.com/crimewatch/' + re.search(r'\boLyr\.src="\./(map.asp\?opt=attrib[^"]+)";', crimeDetailsHTML, re.I).group(1)
        
        self.requestCrimeDetailLoadingFrame(iframeURL, crimeRequest)

    def requestCrimeDetailLoadingFrame(self, crimeDetailLoadingFrameURL, crimeRequest):
        """ Request HTML source of crime match detail loading frame.
        """
        print '-' * 64
        print 'Sending crime detail loading frame request:', str(crimeRequest)
        print self.requestStatus()

        agent = self

        def onRequestCrimeDetailLoadingFrame(client):
            reactor.callLater(0, agent.onRequestCrimeDetailLoadingFrame,
                                 client.response,
                                 crimeRequest)
                  
        self.makeRequest(onRequestCrimeDetailLoadingFrame,
                         crimeDetailLoadingFrameURL,
                         headers={'Referer': 'http://gismaps.oaklandnet.com/crimewatch/map.asp'},
                         cookies=crimeRequest.mapRequest.cookieJar, agent=self.userAgent)
        
    def onRequestCrimeDetailLoadingFrame(self, crimeDetailLoadingFrameHTML, crimeRequest):
        """ HTML source of crime match detail loading frame has been retrieved.
        """
        print '-' * 64
        print 'Crime details frame HTML:', len(crimeDetailLoadingFrameHTML), 'bytes'
        print 'Request:', str(crimeRequest)
        
        # yet another iframe URL deep inside a javascript statement
        iframeURL = 'http://gismaps.oaklandnet.com/crimewatch/' + re.search(r"\bwindow.location.replace\('./(map.asp\?opt=attrib[^']+)'\);", crimeDetailLoadingFrameHTML, re.I).group(1)

        self.requestCrimeDetailDataFrame(iframeURL, crimeRequest)

    def requestCrimeDetailDataFrame(self, crimeDetailDataFrameURL, crimeRequest):
        """ Request HTML source of crime match detail data frame.
        """
        print '-' * 64
        print 'Sending crime detail data frame request:', str(crimeRequest)
        print self.requestStatus()

        agent = self

        def onRequestCrimeDetailDataFrame(client):
            reactor.callLater(0, agent.onRequestCrimeDetailDataFrame,
                                 BeautifulSoup.BeautifulSoup(client.response),
                                 crimeRequest)
                  
        self.makeRequest(onRequestCrimeDetailDataFrame,
                         crimeDetailDataFrameURL,
                         headers={'Referer': 'http://gismaps.oaklandnet.com/crimewatch/map.asp'},
                         cookies=crimeRequest.mapRequest.cookieJar, agent=self.userAgent)
        
    def onRequestCrimeDetailDataFrame(self, crimeDetailDataFrameTagSoup, crimeRequest):
        """ HTML source of crime match detail data frame has been retrieved.
        """
        print '-' * 64
        print 'Crime details frame HTML soup:', len(str(crimeDetailDataFrameTagSoup)), 'bytes'
        print 'Request:', str(crimeRequest)
        
        self.extractCrimeDetails(crimeDetailDataFrameTagSoup, crimeRequest)
        
    def extractCrimeDetails(self, crimeSoup, crimeRequest):
        """ Extract crime details from HTML tag soup.
        """
        try:
            def enclosingTR(tag):
                return tag.findParent('tr')
            
            crimeDetails = map(None, map(enclosingTR, crimeSoup.fetchText(re.compile('^Crime&nbsp;Type$'))),
                                     map(enclosingTR, crimeSoup.fetchText(re.compile('^Date$'))),
                                     map(enclosingTR, crimeSoup.fetchText(re.compile('^Time$'))),
                                     map(enclosingTR, crimeSoup.fetchText(re.compile('^Case&nbsp;Number$'))),
                                     map(enclosingTR, crimeSoup.fetchText(re.compile('^Description$'))))
            
            for crimeDetail in crimeDetails:
                crimeType   = crimeDetail[0]('td')[-1].contents[0].replace('&nbsp;', ' ')
                dateStr     = crimeDetail[1]('td')[-1].contents[0]
                timeStr     = crimeDetail[2]('td')[-1].contents[0]
                caseNumber  = crimeDetail[3]('td')[-1].contents[0].replace('&nbsp;', ' ')
                description = crimeDetail[4]('td')[-1].contents[0].replace('&nbsp;', ' ')
                
                dateTime = datetime.datetime(*time.strptime(dateStr + timeStr, '%Y%m%d%H%M')[:6])
                location = crimeRequest.district.project(*crimeRequest.crimeMatch.center())

                crime = Store.Crime(crimeRequest, dateTime, location[0], location[1], description, caseNumber)
            
                print '-' * 8
                print 'Crime type:',  crime.crime.id
                print 'Date/time:',   str(crime.dateTime)
                print 'Location:',    (crime.latitude, crime.longitude)
                print 'Case number:', crime.caseNumber
                print 'Description:', crime.description
                
                crime.save(self.dataStore)
                
        except Exception, e:
            print 'Something went wrong while extracting crime details:'
            print e
        
        self.finishRequest(crimeRequest)
        
    def finishRequest(self, request):
        self.completedRequests += 1
        self.requestsInProgress -= 1

        if request.__class__ is CrimeRequest and request.nextRequest():
            self.requestQueue.insert(0, request.nextRequest())
        
        print '-' * 64
        print 'Finished request:', str(request)
        print self.requestStatus()
        
        if (len(self.requestQueue) + self.requestsInProgress) == 0:
            reactor.stop()

if __name__ == '__main__':
    import doctest
    doctest.testmod()
