"""
>>> m1 = Marker(  0,   0, 35, -125)
>>> m2 = Marker(100,   0, 35, -120)
>>> m3 = Marker(  0, 100, 30, -125)

>>> transformLat = linearSolution(m1.x, m1.y, m1.lat, m2.x, m2.y, m2.lat, m3.x, m3.y, m3.lat)
>>> transformLon = linearSolution(m1.x, m1.y, m1.lon, m2.x, m2.y, m2.lon, m3.x, m3.y, m3.lon)

>>> transformLat(0, 0)
35.0
>>> transformLat(100, 0)
35.0
>>> transformLat(0, 100)
30.0
>>> transformLat(100, 100)
30.0
>>> transformLat(50, 50)
32.5
>>> transformLat(200, 200)
25.0

>>> transformLon(0, 0)
-125.0
>>> transformLon(100, 0)
-120.0
>>> transformLon(0, 100)
-125.0
>>> transformLon(100, 100)
-120.0
>>> transformLon(50, 50)
-122.5
>>> transformLon(200, 200)
-115.0
"""

class Marker:

    def __init__(self, x, y, lat, lon, note=None):
        self.x = float(x)
        self.y = float(y)
        self.lat = float(lat)
        self.lon = float(lon)

        self.note = note

class District:

    def __init__(self, name, id, width, height, *markers):
        self.name   = name
        self.id     = id
        self.width  = width
        self.height = height
        
        self.projectLat = linearSolution(markers[0].x, markers[0].y, markers[0].lat,
                                         markers[1].x, markers[1].y, markers[1].lat,
                                         markers[2].x, markers[2].y, markers[2].lat)
        
        self.projectLon = linearSolution(markers[0].x, markers[0].y, markers[0].lon,
                                         markers[1].x, markers[1].y, markers[1].lon,
                                         markers[2].x, markers[2].y, markers[2].lon)
        
    def project(self, x, y):
        return self.projectLat(x, y), self.projectLon(x, y)

def linearSolution(x1, y1, z1, x2, y2, z2, x3, y3, z3):
    """ Solves a system of linear equations.

          z1 = (a * x1) + (b + y1) + c
          z2 = (a * x2) + (b + y2) + c
          z3 = (a * x3) + (b + y3) + c
        
        x1 - z3 are the known values.
        a, b, c are the unknowns to be solved.
        generates a function which will return z for any x, y.
        ask mike about pages 78-79 of his notebook for more info.
    """
    
    # make them all floats
    x1, y1, z1, x2, y2, z2, x3, y3, z3 = map(float, (x1, y1, z1, x2, y2, z2, x3, y3, z3))

    a = (((z2 - z3) * (y1 - y2)) - ((z1 - z2) * (y2 - y3))) \
      / (((x2 - x3) * (y1 - y2)) - ((x1 - x2) * (y2 - y3)))
    
    b = (((z2 - z3) * (x1 - x2)) - ((z1 - z2) * (x2 - x3))) \
      / (((y2 - y3) * (x1 - x2)) - ((y1 - y2) * (x2 - x3)))
               
    c = z1 - (x1 * a) - (y1 * b)
		
    return lambda x, y: ((a * x) + (b * y) + c)

def districts():
    return {'COUNCIL DISTRICT 1':   District('Council District 1', 'COUNCIL DISTRICT 1',
                                             2000, 1000,
                                             Marker(  72, 138, 37.874515, -122.324824, 'Tip of Berkeley Marina'),
                                             Marker(1391, 919, 37.821658, -122.207268, 'Park Blvd. & 13'),
                                             Marker( 879, 723, 37.834429, -122.252373, 'Broadway & 51st')),
            'COUNCIL DISTRICT 2':   District('Council District 2', 'COUNCIL DISTRICT 2',
                                             2000, 1000,
                                             Marker(1719,  15, 37.821658, -122.207268, 'Park Blvd. & 13'),
                                             Marker(1904, 794, 37.786818, -122.195896, 'High St. & 580'),
                                             Marker( 362, 125, 37.815548, -122.283980, 'Adeline & W Grand')),
            'COUNCIL DISTRICT 3':   District('Council District 3', 'COUNCIL DISTRICT 3',
                                             2000, 1000,
                                             Marker( 466, 607, 37.805775, -122.342656, 'Near Portview Park'),
                                             Marker(1861, 862, 37.793211, -122.240968, '14th St. & E 20th'),
                                             Marker(1719, 132, 37.834429, -122.252373, 'Broadway & 51st St.')),
            'COUNCIL DISTRICT 4':   District('Council District 4', 'COUNCIL DISTRICT 4',
                                             2000, 1000,
                                             Marker(1801, 830, 37.782824, -122.115419, 'Peninsula on Upper San Leandro Reservoir'),
                                             Marker( 216, 941, 37.771436, -122.276856, 'Central & Webster (Alameda)'),
                                             Marker( 468, 165, 37.834429, -122.252373, 'Broadway & 51st')),
            'COUNCIL DISTRICT 5':   District('Council District 5', 'COUNCIL DISTRICT 5',
                                             2000, 1000,
                                             Marker( 232,  53, 37.809547, -122.286190, 'Adeline & 14th St.'),
                                             Marker(1354,  54, 37.810598, -122.197688, 'Lincoln & 13'),
                                             Marker(1755, 830, 37.762683, -122.164900, '82nd Ave. & MacArthur')),
            'COUNCIL DISTRICT 6':   District('Council District 6', 'COUNCIL DISTRICT 6',
                                             2000, 1000,
                                             Marker(1815, 433, 37.782824, -122.115419, 'Peninsula on Upper San Leandro Reservoir'),
                                             Marker(1290, 941, 37.753717, -122.151264, 'Golf Links Rd. & 580'),
                                             Marker( 142,  18, 37.804291, -122.233415, 'Park Blvd. & 580')),
            'COUNCIL DISTRICT 7':   District('Council District 7', 'COUNCIL DISTRICT 7',
                                             2000, 1000,
                                             Marker(1752, 687, 37.729767, -122.138507, 'Estudillo & 580'),
                                             Marker( 868, 896, 37.715552, -122.209468, 'Airport loop'),
                                             Marker( 105,  13, 37.771325, -122.272800, 'Central Ave. & 8th St. (Alameda)')),
            'PIEDMONT':             District('Piedmont', 'PIEDMONT',
                                             2000, 1000,
                                             Marker(1807, 540, 37.821658, -122.207268, 'Park Blvd. & 13'),
                                             Marker( 146,  79, 37.831251, -122.254422, 'Broadway & 42nd St.'),
                                             Marker( 420, 843, 37.814192, -122.246214, 'Grand & Mandana'))}

if __name__ == '__main__':
    import doctest
    doctest.testmod()
    