import operator as op, math
import PIL.Image as Image
import PIL.ImageChops as IChops
import PIL.ImageFilter as IFilter
import PIL.ImageDraw as IDraw
import PIL.ImageStat as IStat
import Numeric as Num

def floor(x):
    return int(math.floor(x))

def ceil(x):
    return int(math.ceil(x))

def arr2img(ar):
    """ Convert Numeric.array to PIL.Image.
    """
    return Image.fromstring('L', (len(ar), len(ar[0])), ar.astype('b').tostring())

def img2arr(im):
    """ Convert PIL.Image to Numeric.array.
    """
    return Num.reshape(Num.fromstring(im.tostring(), 'b'), im.size)

class Match:
    """ A possible match for an icon on the map.
    """
    def __init__(self, crime, box, area, part, distance, threshold):
        self.crime     = crime
        self.box       = box       # (left, top, right, bottom)
        self.area      = area      # this is a small, lazy-copied PIL.Image instance
        self.part      = part      # one of Icon.RIGHT, etc.
        self.distance  = distance
        self.threshold = threshold
        
    def __str__(self):
        crime     = self.crime.name
        icon_part = self.part
        strength  = (1.0 - (float(self.distance) / float(self.threshold))) * 100
        center    = self.center()

        return '%(crime)s, %(icon_part)s %(strength)d%% %(center)s' % locals()
        
    def center(self):
        x = self.box[0] + .5 * (self.box[2] - self.box[0])
        y = self.box[1] + .5 * (self.box[3] - self.box[1])
        return x, y

class Icon:
    """ An icon with some geometry and channel information pre-computed.
    """
    WHOLE = 'Entire icon'
    TOP = 'Top half'
    LEFT = 'Left half'
    RIGHT = 'Right half'
    BOTTOM = 'Bottom half'
    TOP_LEFT = 'Top-left half'
    TOP_RIGHT = 'Top-right half'
    BOTTOM_LEFT = 'Bottom-left half'
    BOTTOM_RIGHT = 'Bottom-right half'

    def __init__(self, filename):
        self.src = filename
        self.img = Image.open(self.src).convert('RGBA')

        self.W = self.img.size[0]
        self.H = self.img.size[1]

        self.R, self.G, self.B, self.A = map(img2arr, self.img.split())
    
class Crime:
    """ A type of crime with hints on how best to visually match its associated icon.
    """
    AND = 'intersection of colors'
    OR = 'union of colors'

    def __init__(self, name, id, icon, colors, choke, spread, combine=AND):
        self.name = name
        self.id = id
        self.icon = icon
        self.colors = colors
        self.choke = choke
        self.spread = spread
        self.combine = combine

def crimes():
    """ Each type of crime has an associated icon and visual match characteristics.
    """
    return {'AGGRAVATED ASSAULT':   Crime('Aggravated Assault', 'AGGRAVATED ASSAULT',
                                          Icon('icons/aggravated_assault.png'),
                                          [(0xE5, 0x05, 0x08)], 3, 5),
            'ALCOHOL':              Crime('Alcohol', 'ALCOHOL',
                                          Icon('icons/alcohol.png'),
                                          [(0x6B, 0x06, 0x08)], 3, 7),
            'ARSON':                Crime('Arson', 'ARSON',
                                          Icon('icons/arson.png'),
                                          [(0xEF, 0x08, 0x0C)], 3, 7),
            'BURGLARY':             Crime('Burglary', 'BURGLARY',
                                          Icon('icons/burglary.png'),
                                          [(0x05, 0x03, 0x05), (0xFD, 0xFC, 0xEF)], 1, 5, Crime.AND),
            'DISTURBING THE PEACE': Crime('Disturbing the Peace', 'DISTURBING THE PEACE',
                                          Icon('icons/disturbing_the_peace.png'),
                                          [(0x04, 0x03, 0x05)], 3, 5),
            'MURDER':               Crime('Murder', 'MURDER',
                                          Icon('icons/murder.png'),
                                          [(0xE2, 0x0C, 0x0F), (0xB6, 0x20, 0x21)], 3, 5, Crime.OR),
            'NARCOTICS':            Crime('Narcotics', 'NARCOTICS',
                                          Icon('icons/narcotics.png'),
                                          [(0x10, 0x68, 0xE9), (0x45, 0x68, 0x7C)], 1, 5, Crime.OR),
            'ROBBERY':              Crime('Robbery', 'ROBBERY',
                                          Icon('icons/robbery.png'),
                                          [(0x05, 0x03, 0x07), (0xFD, 0xFB, 0xEF)], 1, 5, Crime.AND),
            'PROSTITUTION':         Crime('Prostitution', 'PROSTITUTION',
                                          Icon('icons/prostitution.png'),
                                          [(0xF7, 0x05, 0xC4), (0xD0, 0x1E, 0xAA)], 1, 5, Crime.OR),
            'SIMPLE ASSAULT':       Crime('Simple Assault', 'SIMPLE ASSAULT',
                                          Icon('icons/simple_assault.png'),
                                          [(0x05, 0x04, 0xE0)], 3, 5),
            'THEFT':                Crime('Theft', 'THEFT',
                                          Icon('icons/theft.png'),
                                          [(0x45, 0x9D, 0x1D)], 1, 5),
            'VANDALISM':            Crime('Vandalism', 'VANDALISM',
                                          Icon('icons/vandalism.png'),
                                          [(0x8D, 0x17, 0xBB)], 1, 7),
            'VEHICLE THEFT':        Crime('Vehicle Theft', 'VEHICLE THEFT',
                                          Icon('icons/vehicle_theft.png'),
                                          [(0xF2, 0x57, 0x0F), (0xFD, 0xF2, 0xCD)], 1, 7, Crime.AND)}

def getMatch(img, x, y, crime, part=Icon.WHOLE):
    """ Determine whether an icon exists at a given x, y location in an image.
    """
    box = ((x - ceil(crime.icon.W/2.)), (y - ceil(crime.icon.H/2.)), (x + floor(crime.icon.W/2.)), (y + floor(crime.icon.H/2.)))
    
    # area of interest from the complete image
    area = img.crop(box)
    areaR, areaG, areaB = map(img2arr, area.split())

    # alpha mask of meaningful icon pixels
    alpha = crime.icon.A.copy()
    
    if part == Icon.TOP:
        # interested only in the top; bottom of alpha channel goes to black
        alpha[:,floor(area.size[1]*.35):] = 0

    elif part == Icon.LEFT:
        # interested only in the left; right of alpha channel goes to black
        alpha[floor(area.size[0]*.35):,:] = 0

    elif part == Icon.RIGHT:
        # interested only in the right; left of alpha channel goes to black
        alpha[:ceil(area.size[0]*.65),:] = 0

    elif part == Icon.BOTTOM:
        # interested only in the bottom; top of alpha channel goes to black
        alpha[:,:ceil(area.size[1]*.65)] = 0

    # determine how different the image is from the icon
    diffR = Num.absolute(crime.icon.R.astype('l') - areaR.astype('l'))
    diffG = Num.absolute(crime.icon.G.astype('l') - areaG.astype('l'))
    diffB = Num.absolute(crime.icon.B.astype('l') - areaB.astype('l'))
    diffSum = (diffR + diffG + diffB)
    diffMask = alpha * 255 * 3
    diffFinal = Num.minimum(diffSum, diffMask)
    
    distance = sum(sum(diffFinal))
    threshold = 2 * 32 * sum(sum(alpha / 255))
    
    if distance < threshold:
        # close enough
        match = Match(crime, box, area, part, distance, threshold)
        print '  Found:', match
        return match
        
    elif distance < (2 * threshold) and part == Icon.WHOLE:
        # close enough to check for a partially-occluded icon

        top = getMatch(img, x, y, crime, Icon.TOP)
        if top:
            return top
        
        left = getMatch(img, x, y, crime, Icon.LEFT)
        if left:
            return left
        
        right = getMatch(img, x, y, crime, Icon.RIGHT)
        if right:
            return right
        
        bottom = getMatch(img, x, y, crime, Icon.BOTTOM)
        if bottom:
            return bottom
        
    return None
        
def getMatches(img, mask, crime, box=None):
    """ Search for possible crime icon instances in an image, with a mask of possible locations.
    """
    matches = []

    if box is None:
        # start at the whole box
        box = (0, 0, img.size[0], img.size[1])

    width  = box[2] - box[0]
    height = box[3] - box[1]

    hits = mask.crop(box).histogram()[-1]
    
    if hits > 0:
        # there exists a candidate somewhere in the current box

        if (width == 1) or (height == 1):
            # the box is very small, or at least very thin - check each pixel for a match
            for x in range(box[0], box[2]):
                for y in range(box[1], box[3]):
                    if mask.getpixel((x, y)):
                        match = getMatch(img, x, y, crime)
                        if match is not None:
                            matches.append(match)

        else:
            # break the box down into quadrants, and recursively search each one
            boxTL = (box[0],         box[1],          box[0]+width/2, box[1]+height/2)
            boxTR = (box[0]+width/2, box[1],          box[2],         box[1]+height/2)
            boxBL = (box[0],         box[1]+height/2, box[0]+width/2, box[3])
            boxBR = (box[0]+width/2, box[1]+height/2, box[2],         box[3])
            
            matches += getMatches(img, mask, crime, boxTL)
            matches += getMatches(img, mask, crime, boxTR)
            matches += getMatches(img, mask, crime, boxBL)
            matches += getMatches(img, mask, crime, boxBR)

    return matches

def getMask(img, imgR, imgG, imgB, crime):
    """ Mask out unlikely image locations for crime icons.
    
        Returns mask image with black = no icon, white = possible icon.
    """
    # start with an empty list for mask images
    masks = []
    
    for R, G, B in crime.colors:
        # calculate difference between image and characteristic colors for crime icon
        diffR = Num.absolute(imgR - (Num.ones(img.size) * R))
        diffG = Num.absolute(imgG - (Num.ones(img.size) * G))
        diffB = Num.absolute(imgB - (Num.ones(img.size) * B))
        diffD = diffR + diffG + diffB

        mask = Num.less(diffD, Num.ones(img.size) * 48) * 255
        masks.append(arr2img(mask))
        
    if crime.combine == Crime.OR:
        # combine mask channels into one image
        mask = reduce(IChops.lighter, masks)
    
        # choke out stray pixels, inflate others
        mask = mask.filter(IFilter.MinFilter(crime.choke))
        mask = mask.filter(IFilter.MaxFilter(crime.spread))
        
    else:
        # choke out stray pixels, inflate others
        masks = [mask.filter(IFilter.MinFilter(crime.choke)) for mask in masks]
        masks = [mask.filter(IFilter.MaxFilter(crime.spread)) for mask in masks]
        
        # combine mask channels into one image
        mask = reduce(IChops.darker, masks)
        
    return mask

if __name__ == '__main__':
    import doctest
    doctest.testmod()
