"""
Set up a testing DB:

>>> db = getDB(':memory:')

>>> db.execute(\"""CREATE TABLE crimes (
...                 id          INTEGER PRIMARY KEY AUTOINCREMENT,
...
...                 crime_type  TEXT,
...                 date_time   TEXT,
...                 latitude    FLOAT,
...                 longitude   FLOAT,
...                 description TEXT,
...                 case_number TEXT,
...                 request     TEXT,
...
...                 UNIQUE (case_number, description)
...             )\""") # doctest: +ELLIPSIS

Save a crime:

>>> c = Image.crimes()['ARSON']
>>> d = Geocode.districts()['COUNCIL DISTRICT 1']
>>> m = Image.Match(c, (16, 16, 32, 32), None, Image.Icon.WHOLE, 100, 200)
>>> r = Agent.MapRequest('Council_District_Renderer', d.id, '20061116', '20061116', c.id)
>>> Crime(Agent.CrimeRequest(m, r, d),
...       datetime.datetime(2006, 11, 16, 12, 0, 0),
...       37.0, -122.0,
...       'NONEXISTENT CRIME',
...       '01-000000').save(db)

Save another:

>>> c = Image.crimes()['SIMPLE ASSAULT']
>>> d = Geocode.districts()['COUNCIL DISTRICT 3']
>>> m = Image.Match(c, (16, 16, 32, 32), None, Image.Icon.WHOLE, 100, 200)
>>> r = Agent.MapRequest('Council_District_Renderer', d.id, '20061116', '20061116', c.id)
>>> Crime(Agent.CrimeRequest(m, r, d),
...       datetime.datetime(2006, 11, 16, 12, 0, 0),
...       37.0, -122.0,
...       'ANOTHER NONEXISTENT CRIME',
...       '01-000001').save(db)

Read out the first two:

>>> for crime in db.query('SELECT * FROM crimes ORDER BY id ASC'):
...     print crime
(1, u'ARSON', u'2006-11-16 12:00:00', 37.0, -122.0, u'NONEXISTENT CRIME', u'01-000000', u'Arson @24,24 (COUNCIL DISTRICT 1, ARSON: 20061116 - 20061116), 0 remaining')
(2, u'SIMPLE ASSAULT', u'2006-11-16 12:00:00', 37.0, -122.0, u'ANOTHER NONEXISTENT CRIME', u'01-000001', u'Simple Assault @24,24 (COUNCIL DISTRICT 3, SIMPLE ASSAULT: 20061116 - 20061116), 0 remaining')

Save a third, that should actually overwrite the first:

>>> c = Image.crimes()['AGGRAVATED ASSAULT']
>>> d = Geocode.districts()['COUNCIL DISTRICT 1']
>>> m = Image.Match(c, (16, 16, 32, 32), None, Image.Icon.WHOLE, 100, 200)
>>> r = Agent.MapRequest('Council_District_Renderer', d.id, '20061116', '20061116', c.id)
>>> Crime(Agent.CrimeRequest(m, r, d),
...       datetime.datetime(2006, 11, 16, 12, 0, 0),
...       37.0, -122.0,
...       'NONEXISTENT CRIME',
...       '01-000000').save(db)

See if the replacement worked correctly:

>>> for crime in db.query('SELECT * FROM crimes ORDER BY id ASC'):
...     print crime
(2, u'SIMPLE ASSAULT', u'2006-11-16 12:00:00', 37.0, -122.0, u'ANOTHER NONEXISTENT CRIME', u'01-000001', u'Simple Assault @24,24 (COUNCIL DISTRICT 3, SIMPLE ASSAULT: 20061116 - 20061116), 0 remaining')
(3, u'AGGRAVATED ASSAULT', u'2006-11-16 12:00:00', 37.0, -122.0, u'NONEXISTENT CRIME', u'01-000000', u'Aggravated Assault @24,24 (COUNCIL DISTRICT 1, AGGRAVATED ASSAULT: 20061116 - 20061116), 0 remaining')

"""
import datetime
import Image, Agent, Geocode
from pysqlite2 import dbapi2 as sqlite

class DBWrapper:
    def __init__(self, cursor):
        self.cursor = cursor

    def query(self, *args, **kwargs):
        self.cursor.execute(*args, **kwargs)
        return self.cursor.fetchall()

    def execute(self, *args, **kwargs):
        self.cursor.execute(*args, **kwargs)

class Crime:

    def __init__(self, request, dateTime, latitude, longitude, description, caseNumber):
        self.request     = request              # Agent.CrimeRequest
        self.match       = request.crimeMatch   # Image.Match
        self.crime       = self.match.crime     # Image.Crime
        self.dateTime    = dateTime             # datetime.datetime
        self.latitude    = latitude
        self.longitude   = longitude
        self.description = description
        self.caseNumber  = caseNumber
        
    def save(self, db):
        crimeType = self.crime.id
        dateTime = str(self.dateTime)
        latitude, longitude = self.latitude, self.longitude
        description = self.description
        caseNumber = self.caseNumber
        request = str(self.request)
    
        db.execute("""INSERT OR REPLACE INTO crimes
                      (crime_type, date_time, latitude, longitude, description, case_number, request)
                      VALUES(?, ?, ?, ?, ?, ?, ?)""",
                   (crimeType, dateTime, latitude, longitude, description, caseNumber, request))

def getDB(source):
    return DBWrapper(sqlite.connect(source, isolation_level=None).cursor())

if __name__ == '__main__':
    import doctest
    doctest.testmod()
